#include "EspMQTTClient.h"
#include <CommandParser.h>
#include <TimeOut.h>

//MQTT Client Configuration
EspMQTTClient client(
  "Your_SSID",
  "Your_WIFI_PWD",
  "192.168.3.33",  // MQTT Broker server ip
  "",   // Can be omitted if not needed
  "",   // Can be omitted if not needed
  "SmartAgr",     // Client name that uniquely identify your device
  1883              // The MQTT port, default to 1883. this line can be omitted
);


// Pin Definition
int R1 = 2;     //Relay 1
int R2 = 15;    //Relay 2
int P1 = 4;    //5V Power output 1
int P2 = 16;     //5V Power output 2
int S1= 36;     //Push Button on the board
int AD4=32;     //ADC4 for Soil Moisture Sensor 1

//Command Parser
typedef CommandParser<5, 3, 10, 255, 255> MyCommandParser;
//typedef CommandParser<> MyCommandParser;
MyCommandParser parser;

enum IOstate {R1ON=0, R1OFF=1, R2ON=2, R2OFF=3, P1ON=4, P1OFF=5, P2ON=6, P2OFF=7} ;

//Timeout
TimeOut timeoutR1;
TimeOut timeoutR2;
TimeOut timeoutP1;
TimeOut timeoutP2;
TimeOut timeoutAD4;
//Callback function for SetIO
void callbackSetIO(int IOState);
//Callback function for AD4 (Soil moisture Sensor 1)
#define MeasurementsToAverage 10
#define SampleInterval  60000 //60s
void callbackAD4();

void setup() {
  //Init Realies and 5V Power outputs
  pinMode (R1, OUTPUT);
  pinMode (R2, OUTPUT);
  pinMode (P1, OUTPUT);
  pinMode (P2, OUTPUT);

  //Init Button
  pinMode(S1,INPUT);

  //Init Serial Port
  Serial.begin(115200);

  //Init MQTT Client
  // Optionnal functionnalities of EspMQTTClient : 
  client.enableDebuggingMessages(); // Enable debugging messages sent to serial output
  client.enableHTTPWebUpdater("neil","neilhao"); // Enable the web updater. User and password default to values of MQTTUsername and MQTTPassword. These can be overrited with enableHTTPWebUpdater("user", "password").
  client.enableLastWillMessage("EndDevices/lastwill", "I am going offline");  // You can activate the retain flag by setting the third parameter to true

  //Reg Commands
   parser.registerCommand("ATS+", "su", &mqtt_cmd);
  
  //Update the Soil moisture Sensor 1 Data to MQTT every SampleInterval s
  timeoutAD4.timeOut(SampleInterval,callbackAD4);
}

// This function is called once everything is connected (Wifi and MQTT)
// WARNING : YOU MUST IMPLEMENT IT IF YOU USE EspMQTTClient
void onConnectionEstablished()
{
  // Subscribe to "mytopic/test" and display received message to Serial
  client.subscribe("HOME/SA01/IO", [](const String & payload) {
    Serial.println(payload);

    char response[MyCommandParser::MAX_RESPONSE_SIZE];
    char line[128];
    payload.toCharArray(line, payload.length()+1);
    parser.processCommand(line, response);
    Serial.println(response);
  });
}

void loop() {
  int S1_Button_State = digitalRead(S1);
  if(S1_Button_State == LOW){ //The S1 Button is pushing down, entering IO testing mode
  digitalWrite(R1, HIGH);
  digitalWrite(R2, HIGH);
  digitalWrite(P1, HIGH);
  digitalWrite(P2, HIGH);
  delay(1000);
  digitalWrite(R1, LOW);
  digitalWrite(R2, LOW);
  digitalWrite(P1, LOW);
  digitalWrite(P2, LOW);
  delay(1000);
  }

  client.loop();
  //monitor handler, Have to be into void loop
  //handler is static. you can call:
  TimeOut::handler();
}

void mqtt_cmd(MyCommandParser::Argument *args, char *response) {
  Serial.print("string: "); Serial.println(args[0].asString);
  Serial.print("uint64: "); Serial.println((uint32_t)args[1].asUInt64); // NOTE: on older AVR-based boards, Serial doesn't support printing 64-bit values, so we'll cast it down to 32-bit

  if((String)args[0].asString == "R1ON"){
    Serial.println("Turn on R1");
    //Trun on Relay 1
    SetIO(R1ON);
    //Trun off Relay 1 after (uint32_t)args[1].asUInt64 ms
    timeoutR1.timeOut<int>((uint32_t)args[1].asUInt64, callbackSetIO, R1OFF );
  }

  if((String)args[0].asString == "R2ON"){
    Serial.println("Turn on R2");
    //Trun on Relay 2
    SetIO(R2ON);
    //Trun off Relay 2 after (uint32_t)args[1].asUInt64 ms
    timeoutR2.timeOut<int>((uint32_t)args[1].asUInt64, callbackSetIO, R2OFF );
  }

  if((String)args[0].asString == "P1ON"){
    Serial.println("Turn on P1");
    //Trun on Power out 1
    SetIO(P1ON);
    //Trun off Power out 1 after (uint32_t)args[1].asUInt64 ms
    timeoutP1.timeOut<int>((uint32_t)args[1].asUInt64, callbackSetIO, P1OFF );
  }

  if((String)args[0].asString == "P2ON"){
    Serial.println("Turn on P2");
    //Trun on Power out 2
    SetIO(P2ON);
    //Trun off Power out 2 after (uint32_t)args[1].asUInt64 ms
    timeoutP2.timeOut<int>((uint32_t)args[1].asUInt64, callbackSetIO, P2OFF );
  }


//Turn Off
  if((String)args[0].asString == "R1OFF"){
    Serial.println("Turn off R1");
    //Trun off Relay 1
    SetIO(R1OFF);
    timeoutR1.cancel();
  }

  if((String)args[0].asString == "R2OFF"){
    Serial.println("Turn off R2");
    //Trun off Relay 2
    SetIO(R2OFF);
    timeoutR2.cancel();
  }

  if((String)args[0].asString == "P1OFF"){
    Serial.println("Turn off P1");
    //Trun off Power out 1
    SetIO(P1OFF);
    timeoutP1.cancel();
  }

  if((String)args[0].asString == "P2OFF"){
    Serial.println("Turn off P2");
    //Trun off Power out 2
    SetIO(P2OFF);
    timeoutP2.cancel();
  }

  strlcpy(response, "success", MyCommandParser::MAX_RESPONSE_SIZE);
}

void SetIO(int IOState){
  switch(IOState){
    case R1ON:
      digitalWrite(R1, HIGH);
      //Feedback
      client.publish("HOME/SA01/IO/FB", "R1ON");
      break;
    case R1OFF:
      digitalWrite(R1, LOW);
      //Feedback
      client.publish("HOME/SA01/IO/FB", "R1OFF");
      break;
    case R2ON:
      digitalWrite(R2, HIGH);
      //Feedback
      client.publish("HOME/SA01/IO/FB", "R2ON");
      break;
    case R2OFF:
      digitalWrite(R2, LOW);
      //Feedback
      client.publish("HOME/SA01/IO/FB", "R2OFF");
      break;
    case P1ON:
      digitalWrite(P1, HIGH);
      //Feedback
      client.publish("HOME/SA01/IO/FB", "P1ON");
      break;
    case P1OFF:
      digitalWrite(P1, LOW);
      //Feedback
      client.publish("HOME/SA01/IO/FB", "P1OFF");
      break;
    case P2ON:
      digitalWrite(P2, HIGH);
      //Feedback
      client.publish("HOME/SA01/IO/FB", "P2ON");
      break;
    case P2OFF:
      digitalWrite(P2, LOW);
      //Feedback
      client.publish("HOME/SA01/IO/FB", "P2OFF");
      break;
   default :
    Serial.println("Error: IO did not find.");
  }

}


void callbackSetIO(int IOState){
  SetIO(IOState);
}

void callbackAD4(){
  uint32_t AverageAD4 = 0;
  //uint32_t MeasurementsToAverage = 10;
  for(int i = 0; i < MeasurementsToAverage; i++)
  {
    AverageAD4 += analogRead(AD4);
    delay(5);
  }
  AverageAD4 = (uint32_t)(AverageAD4/MeasurementsToAverage);
  //Feedback
  client.publish("HOME/SA01/SENSOR/SOIL1/FB", String(AverageAD4));
  //Update the Soil moisture Sensor 1 Data to MQTT every SampleInterval s
  timeoutAD4.timeOut(SampleInterval,callbackAD4);
}